#ifndef __CAMERA_H__
#define __CAMERA_H__

#include <GL/glu.h>
#include <QtOpenGL>
#include <Geometry.h>

/**
 * Camera
 * @author Stphane GINIER
 */
class Camera
{

public:
    enum CameraMode{SPHERICAL,PLANE,FREELOOK};
    Camera();
    ~Camera();
    void setViewport(const QSizeF &viewport);
    float getZoom() const;
    const QSizeF &getViewport() const;
    const QMatrix4x4 &getViewMatrix() const;
    const QMatrix4x4 &getProjectionMatrix() const;
    void setMoveX(int direction);
    void setMoveZ(int direction);
    void setGlobalScale(float scale);
    bool isMoving() const;
    void setZoom(float zoom);
    CameraMode getMode() const;
    void setMode(CameraMode mode);

    void start(const QPointF& mouseXY);
    void rotate(const QPointF& mouseXY);
    void translate(const QPointF& mouseDiff);
    void updateView();
    void updateProjection();
    void zoom(float delta);
    void updateTranslation();
    void loadProjectionMatrix();
    void loadViewMatrix();
    void reset();

private:
    CameraMode mode_; //camera mode
    QQuaternion rot_; //quaternion
    QMatrix4x4 viewMatrix_; //view matrix
    QMatrix4x4 projectionMatrix_; //projection matrix
    QPointF normalizedMouseXY_; //mouse position (between 0 and 1)
    QSizeF viewport_; //viewport size
    float transX_; //x translation
    float transY_; //y translation
    float zoom_; //zoom value
    int moveX_; //free look (strafe)
    int moveZ_; //free look (strafe)
    QVector3D position_; //position (freelook)
    QVector3D center_;  //point where we are looking (freelook)
    QVector3D up_;  //up vector (freelook)
    float theta_; //x mouse angle (freelook)
    float phi_; //y mouse angle (freelook)
    float globalScale_; //solve scale issue
};

#endif /*__CAMERA_H__*/
