#ifndef __PICKING_H__
#define __PICKING_H__

#include <limits>
#include "Mesh.h"
#include "Geometry.h"
#include "Topology.h"

/**
 * Picking
 * @author Stphane GINIER
 */
class Picking
{

public:
    Picking();
    ~Picking();
    void setMesh(Mesh *mesh);
    Mesh* getMesh();
    void setIntersectionPoint(const QVector3D &intersectionPoint);
    const QVector3D& getIntersectionPoint() const;
    int getSelectedTriangle() const;
    std::vector<int>& getSelectedVertices();
    void setRadiusWorldSquared(float radiusWorldSquared);
    float getRadiusWorldSquared() const;
    void setRadiusScreen(float radius);
    float getRadiusScreen() const;

    void intersectionRayMesh(Mesh *mesh,const QVector3D &vertexNear, const QVector3D &vertexFar,
                             const QPointF &mouseXY, const QSizeF &viewport);
    void setPickedVerticesInsideSphere(float radiusWorldSquared);
    void forceCenter(Mesh *mesh, const QVector3D &center, const QPointF &mouseXY, const QSizeF &viewport);

private:
    Mesh *mesh_; //selected mesh
    int pickedTriangle_; //triangle picked
    std::vector<int> pickedVertices_; //vertices selected
    QVector3D intersectionPoint_; //intersection point
    float radiusScreen_; //radius of the selection area (screen unit)
    float radiusWorldSquared_; //radius of the selection area (world unit)
};

#endif /*__PICKING_H__*/
