#ifndef __SCULPT_H__
#define __SCULPT_H__

#include "Mesh.h"
#include "Topology.h"

/**
 * Sculpt
 * @author Stphane GINIER
 */
class Sculpt
{

public:
    enum SculptMode{INFLATE, DEFLATE, SMOOTH, FLATTEN, SWEEP};
    enum TopoMode{DECIMATION, SUBDIVISION, UNIFORMISATION, ADAPTIVE, STATIC};
    Sculpt();
    ~Sculpt();
    void setMesh(Mesh *mesh);
    void setIntensity(float intensity);
    void toggleCulling();
    void setSculptMode(SculptMode mode);
    void setTopoMode(TopoMode mode);
    void setDetail(float detail);
    const QVector3D& getSweepCenter();
    void setSweepCenter(const QVector3D &center);
    void setSweepDir(const QVector3D &dir);
    bool isSweep();

    void sculptMesh(std::vector<int> &iVertsSelected, const QVector3D &intersectionPoint,
                    float radiusSquared, const QVector3D &eyeDirection);
    void smooth(const std::vector<int> &iVerts, float intensity);
    void smoothFlat(const std::vector<int> &iVerts, float intensity);
    void draw(const std::vector<int> &iVerts, float radiusSquared, float intensity);
    void flatten(const std::vector<int> &iVerts, float radiusSquared, float intensity);
    void smoothNoMp(const std::vector<int> &iVerts, bool flat = false);
    void sweep(const std::vector<int> &iVerts, float radiusSquared, float intensity);

private:
    void setAdaptiveParameters(float radiusSquared);
    QVector3D areaNormal(const std::vector<int> &iVerts);
    QVector3D areaCenter(const std::vector<int> &iVerts);
    void laplacianSmooth(const std::vector<int> &iVerts, std::vector<QVector3D> &smoothVerts);

private:
    Mesh* mesh_; //selected mesh
    float intensity_; //deformation intensity
    SculptMode sculptMode_; //sculpting mode
    TopoMode topoMode_; //topological mode
    QVector3D centerPoint_; //center of deformation
    bool culling_; //if backface culling is enabled for sculpting
    float detail_; //intensity of details
    float thickness_; //thickness
    float d2Min_; //uniform refinement of mesh (min edge length)
    float d2Max_; //uniform refinement of mesh (max edge length)
    float d2Thickness_; //distance between 2 vertices before split/merge
    float d2Move_; //max displacement of vertices per step
    QVector3D sweepCenter_; //center of sweep tool
    QVector3D sweepDir_; //direction of deformation for sweep tool
};

#endif /*__SCULPT_H__*/
