#ifndef __WIDGET3D_H__
#define __WIDGET3D_H__

#include <limits>
#include "Camera.h"
#include "Window.h"
#include "Picking.h"
#include "Sculpt.h"
#include "Files.h"
#include <QTime>

/**
 * Widget3D
 * @author Stphane GINIER
 */
class Widget3D : public QGLWidget
{
    Q_OBJECT

public:
    Widget3D(Window *window);
    ~Widget3D();
    void keyPressEvent(QKeyEvent *event);
    void keyReleaseEvent(QKeyEvent *event);

public slots:
    void setSculptRadius(int radius);
    void setSculptIntensity(int intensity);
    void setSculptMode(int mode);
    void setTopoMode(int mode);
    void setTopoDetail(int detail);
    void setSculptCulling();
    void setSymmetry();
    void setCamera(int mode);
    void setMeshColor();
    void setMeshShader(int mode);
    void loadFile();
    void saveFile();
    void undo();
    void redo();

private:
    QSize minimumSizeHint() const;
    QSize sizeHint() const;
    void resizeGL(int width, int height);
    void initializeGL();
    void paintGL();
    void mousePressEvent(QMouseEvent *event);
    void mouseMoveEvent(QMouseEvent *event);
    void mouseReleaseEvent(QMouseEvent *event);
    void wheelEvent(QWheelEvent *event);
    void timerEvent(QTimerEvent *event);
    void sculptStroke(const QPointF &pos);
    void showFps();
    void updateMeshInformation();

private:
    Mesh *mesh_; //selected mesh
    Camera camera_; //camera
    Picking picking_; //picking
    Sculpt sculpt_; //sculpting tool
    QPointF mouseXY_; //mouse coordinates
    QPointF lastMouseXY_; //mouse coordinates (only updated at certain event)
    Window *window_; //window
    int cameraTimer_; //camera timer id
    QTime fpsTime_; //fps timer
    int fpsCounter_; //fps counter
    bool symmetry_; //symmetry sculpting
    float sumDisplacement_; //displacement sum
    bool sculptStart_; //if the sculpting has started
};

#endif /*__WIDGET3D_H__*/
