#include "Window.h"
#include "Widget3D.h"

/** Constructor */
Window::Window() : widget_(0), sculptDock_(0), meshDock_(0), shaderButtonGroup_(0), colorButton_(0)
{
    setWindowTitle(tr("Sculpt3D"));
    widget_= new Widget3D(this);
    setCentralWidget(widget_);
    initSculptDock();
    initMeshDock();
    initMenu();
}

/** Destructor */
Window::~Window()
{
    delete widget_;
}

/** Getters/Setters */
QPushButton* Window::getColorButton() { return colorButton_; }
QButtonGroup* Window::getShaderButtonGroup() { return shaderButtonGroup_; }

/** Initialize sculpt dock*/
void Window::initSculptDock()
{
    sculptDock_ = new QDockWidget("       Sculting       ", this);
    addDockWidget(Qt::LeftDockWidgetArea, sculptDock_);

    //tools group
    QGroupBox *toolsGroupBox = new QGroupBox("tools");
    QLabel *radiusPanel = new QLabel("Radius :");
    QSlider *radiusSlider = new QSlider(Qt::Horizontal);
    radiusSlider->setRange(0, 200);
    radiusSlider->setValue(50);
    QRadioButton *toolInflate = new QRadioButton("inflate");
    toolInflate->setCheckable(true);
    toolInflate->setChecked(true);
    QRadioButton *toolDeflate = new QRadioButton("deflate");
    toolDeflate->setCheckable(true);
    QRadioButton *toolSmooth = new QRadioButton("smooth");
    toolSmooth->setCheckable(true);
    QRadioButton *toolFlatten = new QRadioButton("flatten");
    toolFlatten->setCheckable(true);
    QRadioButton *toolSweep = new QRadioButton("sweep");
    toolSweep->setCheckable(true);
    QButtonGroup *toolsButtonGroup = new QButtonGroup();
    toolsButtonGroup->addButton(toolInflate, Sculpt::INFLATE);
    toolsButtonGroup->addButton(toolDeflate, Sculpt::DEFLATE);
    toolsButtonGroup->addButton(toolSmooth, Sculpt::SMOOTH);
    toolsButtonGroup->addButton(toolFlatten, Sculpt::FLATTEN);
    toolsButtonGroup->addButton(toolSweep, Sculpt::SWEEP);
    QLabel *intensityPanel = new QLabel("Intensity :");
    QSlider *intensitySlider = new QSlider(Qt::Horizontal);
    intensitySlider->setRange(0, 200);
    intensitySlider->setValue(150);
    QVBoxLayout *toolsLayout = new QVBoxLayout();
    toolsLayout->addWidget(radiusPanel);
    toolsLayout->addWidget(radiusSlider);
    toolsLayout->addWidget(toolInflate);
    toolsLayout->addWidget(toolDeflate);
    toolsLayout->addWidget(toolSmooth);
    toolsLayout->addWidget(toolFlatten);
    toolsLayout->addWidget(toolSweep);
    toolsLayout->addWidget(intensityPanel);
    toolsLayout->addWidget(intensitySlider);
    toolsGroupBox->setLayout(toolsLayout);

    //topology group
    QGroupBox *topologyGroupBox = new QGroupBox("topology");
    QRadioButton *topoStatic = new QRadioButton("static");
    QRadioButton *topoSubdivide = new QRadioButton("subdivide");
    QRadioButton *topoDecimation = new QRadioButton("decimate");
    QRadioButton *topoUniformisation = new QRadioButton("uniform");
    topoUniformisation->setChecked(true);
    QRadioButton *topoAdaptive = new QRadioButton("adaptive");
    QButtonGroup *topoButtonGroup = new QButtonGroup();
    topoButtonGroup->addButton(topoStatic, Sculpt::STATIC);
    topoButtonGroup->addButton(topoSubdivide, Sculpt::SUBDIVISION);
    topoButtonGroup->addButton(topoDecimation, Sculpt::DECIMATION);
    topoButtonGroup->addButton(topoUniformisation, Sculpt::UNIFORMISATION);
    topoButtonGroup->addButton(topoAdaptive, Sculpt::ADAPTIVE);
    QLabel *detailPanel = new QLabel("Detail :");
    QSlider *detailSlider = new QSlider(Qt::Horizontal);
    detailSlider->setRange(0, 200);
    detailSlider->setValue(150);
    QVBoxLayout *topologyLayout = new QVBoxLayout();
    topologyLayout->addWidget(topoStatic);
    topologyLayout->addWidget(topoSubdivide);
    topologyLayout->addWidget(topoDecimation);
    topologyLayout->addWidget(topoUniformisation);
    topologyLayout->addWidget(topoAdaptive);
    topologyLayout->addWidget(detailPanel);
    topologyLayout->addWidget(detailSlider);
    topologyGroupBox->setLayout(topologyLayout);

    //options group
    QGroupBox *optionsGroupBox = new QGroupBox("options");
    QCheckBox *cullingButton = new QCheckBox("culling");
    cullingButton->setCheckable(true);
    QCheckBox *symmetryButton = new QCheckBox("symmetry");
    QVBoxLayout *optionsLayout = new QVBoxLayout();
    optionsLayout->addWidget(cullingButton);
    optionsLayout->addWidget(symmetryButton);
    optionsGroupBox->setLayout(optionsLayout);

    QVBoxLayout *dockLayout = new QVBoxLayout();
    dockLayout->addWidget(toolsGroupBox);
    dockLayout->addWidget(topologyGroupBox);
    dockLayout->addWidget(optionsGroupBox);

    QWidget *sculptDockWidget= new QWidget();
    sculptDockWidget->setLayout(dockLayout);
    sculptDockWidget->setFixedSize(sculptDockWidget->sizeHint());

    sculptDock_->setWidget(sculptDockWidget);
    sculptDock_->setFloating(false);

    connect(radiusSlider, SIGNAL(valueChanged(int)), widget_, SLOT(setSculptRadius(int)));
    connect(toolsButtonGroup, SIGNAL(buttonClicked(int)), widget_, SLOT(setSculptMode(int)));
    connect(intensitySlider, SIGNAL(valueChanged(int)), widget_, SLOT(setSculptIntensity(int)));

    connect(topoButtonGroup, SIGNAL(buttonClicked(int)), widget_, SLOT(setTopoMode(int)));
    connect(detailSlider, SIGNAL(valueChanged(int)), widget_, SLOT(setTopoDetail(int)));

    connect(cullingButton, SIGNAL(clicked()), widget_, SLOT(setSculptCulling()));
    connect(symmetryButton, SIGNAL(clicked()), widget_, SLOT(setSymmetry()));

    connect(topoStatic, SIGNAL(toggled(bool)), detailPanel, SLOT(setDisabled(bool)));
    connect(topoStatic, SIGNAL(toggled(bool)), detailSlider, SLOT(setDisabled(bool)));

    connect(toolSweep, SIGNAL(toggled(bool)), symmetryButton, SLOT(setDisabled(bool)));
}

/** Initialize mesh dock */
void Window::initMeshDock()
{
    meshDock_ = new QDockWidget("            Mesh            ", this);
    addDockWidget(Qt::RightDockWidgetArea, meshDock_);

    QGroupBox *cameraGroupBox = new QGroupBox("camera");
    QRadioButton *trackballSphere = new QRadioButton("trackball sphere");
    QRadioButton *trackballPlane = new QRadioButton("trackball plane");
    QRadioButton *freelook = new QRadioButton("freelook");
    trackballSphere->setChecked(true);
    QButtonGroup *cameraButtonGroup = new QButtonGroup();
    cameraButtonGroup->addButton(trackballSphere, Camera::SPHERICAL);
    cameraButtonGroup->addButton(trackballPlane, Camera::PLANE);
    cameraButtonGroup->addButton(freelook, Camera::FREELOOK);
    QVBoxLayout *cameraLayout = new QVBoxLayout();
    cameraLayout->addWidget(trackballSphere);
    cameraLayout->addWidget(trackballPlane);
    cameraLayout->addWidget(freelook);
    cameraGroupBox->setLayout(cameraLayout);

    QGroupBox *colorGroupBox = new QGroupBox("color");
    colorButton_ = new QPushButton();
    colorButton_->setPalette(QPalette(QColor(120,200,200)));
    QVBoxLayout *colorLayout = new QVBoxLayout();
    colorLayout->addWidget(colorButton_);
    colorGroupBox->setLayout(colorLayout);

    QGroupBox *shaderGroupBox = new QGroupBox("shader");
    QRadioButton *phong = new QRadioButton("phong");
    phong->setChecked(true);
    QRadioButton *wireframe = new QRadioButton("wireframe");
    QRadioButton *transparency = new QRadioButton("transparency");
    QRadioButton *reflection1 = new QRadioButton("reflection 1");
    QRadioButton *reflection2 = new QRadioButton("reflection 2");
    shaderButtonGroup_ = new QButtonGroup();
    shaderButtonGroup_->addButton(phong, Mesh::PHONG);
    shaderButtonGroup_->addButton(wireframe, Mesh::WIREFRAME);
    shaderButtonGroup_->addButton(transparency, Mesh::TRANSPARENCY);
    shaderButtonGroup_->addButton(reflection1, Mesh::REFLECTION_1);
    shaderButtonGroup_->addButton(reflection2, Mesh::REFLECTION_2);
    QVBoxLayout *shaderLayout = new QVBoxLayout();
    shaderLayout->addWidget(phong);
    shaderLayout->addWidget(wireframe);
    shaderLayout->addWidget(transparency);
    shaderLayout->addWidget(reflection1);
    shaderLayout->addWidget(reflection2);
    shaderGroupBox->setLayout(shaderLayout);

    QVBoxLayout *dockLayout = new QVBoxLayout();
    dockLayout->addWidget(cameraGroupBox);
    dockLayout->addWidget(colorGroupBox);
    dockLayout->addWidget(shaderGroupBox);

    QWidget *meshDockWidget = new QWidget();
    meshDockWidget->setLayout(dockLayout);
    meshDockWidget->setFixedSize(meshDockWidget->sizeHint());

    meshDock_->setWidget(meshDockWidget);
    meshDock_->setFloating(false);

    connect(cameraButtonGroup, SIGNAL(buttonClicked(int)), widget_, SLOT(setCamera(int)));
    connect(shaderButtonGroup_, SIGNAL(buttonClicked(int)), widget_, SLOT(setMeshShader(int)));
    connect(colorButton_, SIGNAL(clicked()), widget_, SLOT(setMeshColor()));
}

/** Initialize menu */
void Window::initMenu()
{
    QMenu *menuFile = menuBar()->addMenu("&File");

    QAction *openAction = menuFile->addAction("&Open");
    openAction->setShortcut(QKeySequence("Ctrl+O"));

    QAction *saveAction = menuFile->addAction("&Save");
    saveAction->setShortcut(QKeySequence("Ctrl+S"));

    QAction *exitAction = menuFile->addAction("&Exit");
    exitAction->setShortcut(QKeySequence("Ctrl+Q"));

    QMenu *editMenu = menuBar()->addMenu("&Edit");
    QAction *undoAction = editMenu->addAction("&Undo");
    undoAction->setShortcut(QKeySequence("Ctrl+Z"));
    QAction *redoAction = editMenu->addAction("&Redo");
    redoAction->setShortcut(QKeySequence("Ctrl+Y"));

    QMenu *displayMenu = menuBar()->addMenu("&Display");
    displayMenu ->addAction(sculptDock_->toggleViewAction());
    displayMenu->addAction(meshDock_->toggleViewAction());

    connect(exitAction, SIGNAL(triggered()), qApp, SLOT(quit()));
    connect(saveAction, SIGNAL(triggered()), widget_,SLOT(saveFile()));
    connect(openAction, SIGNAL(triggered()), widget_, SLOT(loadFile()));
    connect(undoAction, SIGNAL(triggered()), widget_, SLOT(undo()));
    connect(redoAction, SIGNAL(triggered()), widget_, SLOT(redo()));
}

/** Event */
void Window::keyPressEvent(QKeyEvent *event) { widget_->keyPressEvent(event); }
void Window::keyReleaseEvent(QKeyEvent *event) { widget_->keyReleaseEvent(event); }
