#ifndef __VERTEX_H__
#define __VERTEX_H__

#include <QVector3D>
#include <QtOpenGL>
#include <vector>
#include "Tools.h"

/**
 * Vertex
 * @author Stphane GINIER
 */
class Vertex : public QVector3D
{
public:
    static int tagMask_; //flag mask value (should be always >= tagFlag_)
    static int sculptMask_; //flag mask value (should be always >= sculptFlag_)
    int tagFlag_; //general purpose flag (<0 means the vertex is to be deleted)
    int sculptFlag_; //sculpting flag
    int stateFlag_; //flag for history

public:
    Vertex(float x = 0, float y = 0, float z = 0, int id = -1);
    Vertex(const QVector3D &vec, int id=-1);
    Vertex& operator=(const QVector3D &vec);
    ~Vertex();
    void addTriangle(int iTri);
    void addRingVertex(int iVer);
    void replaceTriangle(int iTriOld, int iTriNew);
    void replaceRingVertex(int iVerOld, int iVerNew);
    void removeTriangle(int iTri);
    void removeRingVertex(int iVer);

public:
    int id_; //id
    std::vector<int> tIndices_; //triangles indices around the vertex
    QVector3D normal_; //normal
    std::vector<int> ringVertices_; //vertices neighborhood (1-ring)
};

bool operator<(const Vertex &a,const Vertex &b);

#endif /*__VERTEX_H__*/
